/*
 * Copyright (C) 2008 Till Harbaum <till@harbaum.org>.
 *
 * This file is part of OSM2Go.
 *
 * OSM2Go is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OSM2Go is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OSM2Go.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define __USE_XOPEN
#include <time.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "appdata.h"

#ifndef LIBXML_TREE_ENABLED
#error "Tree not enabled in libxml"
#endif

/* determine where a node/way/relation read from the osm file */
/* is inserted into the internal database */
// #define GPX_SORT_ID
#define GPX_SORT_LAST
// #define GPX_SORT_FIRST

/* ------------------------- user handling --------------------- */

static void gpx_bounds_free(bounds_t *bounds) {
  free(bounds);
}

static void gpx_bounds_dump(bounds_t *bounds) {
  printf("\nBounds: %f->%f %f->%f\n", 
	 bounds->ll_min.lat, bounds->ll_max.lat, 
	 bounds->ll_min.lon, bounds->ll_max.lon);

  char str[32];
  utm2str(str, sizeof(str), &bounds->center);
  printf("   -> map center is at %s\n", str);
}

static bounds_t *gpx_parse_osm_bounds(gpx_t *gpx, 
		     xmlDocPtr doc, xmlNode *a_node) {
  char *prop;

  if(gpx->bounds) {
    errorf(NULL, "Doubly defined bounds");
    return NULL;
  }

  bounds_t *bounds = g_new0(bounds_t, 1);

  bounds->ll_min.lat = bounds->ll_min.lon = NAN;
  bounds->ll_max.lat = bounds->ll_max.lon = NAN;

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"minlat"))) {
    bounds->ll_min.lat = g_ascii_strtod(prop, NULL);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"maxlat"))) {
    bounds->ll_max.lat = g_ascii_strtod(prop, NULL);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"minlon"))) {
    bounds->ll_min.lon = g_ascii_strtod(prop, NULL);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"maxlon"))) { 
    bounds->ll_max.lon = g_ascii_strtod(prop, NULL); 
    xmlFree(prop); 
  }

  if(isnan(bounds->ll_min.lat) || isnan(bounds->ll_min.lon) || 
     isnan(bounds->ll_max.lat) || isnan(bounds->ll_max.lon)) {
    errorf(NULL, "Invalid coordinate in bounds (%f/%f/%f/%f)",
	   bounds->ll_min.lat, bounds->ll_min.lon, 
	   bounds->ll_max.lat, bounds->ll_max.lon);

    g_free(bounds);
    return NULL;
  }


  /* calculate map zone which will be used as a reference for all */
  /* drawing/projection later on */
  pos_t center = { (bounds->ll_max.lat + bounds->ll_min.lat)/2, 
		   (bounds->ll_max.lon + bounds->ll_min.lon)/2 };

  utm_pos2utm(&center, &bounds->center, NULL);

  utm_t utm;
  utm_pos2utm(&bounds->ll_min, &utm, &bounds->center);
  bounds->min.x = utm.easting - bounds->center.easting;
  bounds->min.y = utm.northing - bounds->center.northing;
  utm_pos2utm(&bounds->ll_max, &utm, &bounds->center);
  bounds->max.x = utm.easting - bounds->center.easting;
  bounds->max.y = utm.northing - bounds->center.northing;

  return bounds;
}

/* ------------------------- user handling --------------------- */

void gpx_users_free(user_t *user) {
  while(user) {
    user_t *next = user->next;

    if(user->name) g_free(user->name);
    g_free(user);

    user = next;
  }
}

void gpx_users_dump(user_t *user) {
  printf("\nUser list:\n");
  while(user) {
    printf("Name: %s\n", user->name);
    user = user->next;
  }
}

static user_t *gpx_user(gpx_t *gpx, char *name) {

  /* search through user list */
  user_t **user = &gpx->user;
  while(*user && strcasecmp((*user)->name, name) < 0) 
    user = &(*user)->next;

  /* end of list or inexact match? create new user entry! */
  if(!*user || strcasecmp((*user)->name, name)) {
    user_t *new = g_new0(user_t, 1);
    new->name = g_strdup(name);
    new->next = *user;
    *user = new;

    return new;
  }

  return *user;
}

static
time_t convert_iso8601(const char *str) {
  tzset();

  struct tm ctime;
  memset(&ctime, 0, sizeof(struct tm));
  strptime(str, "%FT%T%z", &ctime);
 
  return mktime(&ctime) - timezone;
}

/* -------------------- tag handling ----------------------- */

void gpx_tag_free(tag_t *tag) {
  if(tag->key)   g_free(tag->key);
  if(tag->value) g_free(tag->value);
  g_free(tag);
}

static void gpx_tags_free(tag_t *tag) {
  while(tag) {
    tag_t *next = tag->next;
    gpx_tag_free(tag);
    tag = next;
  }
}

static void gpx_tags_dump(tag_t *tag) {
  while(tag) {
    printf("Key/Val: %s/%s\n", tag->key, tag->value);
    tag = tag->next;
  }
}

tag_t *gpx_parse_osm_tag(gpx_t *gpx, xmlDocPtr doc, xmlNode *a_node) {
  xmlNode *cur_node = NULL;

  /* allocate a new tag structure */
  tag_t *tag = g_new0(tag_t, 1);

  char *prop;
  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"k"))) {
    if(strlen(prop) > 0) tag->key = g_strdup(prop);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"v"))) {
    if(strlen(prop) > 0) tag->value = g_strdup(prop);
    xmlFree(prop);
  }

  if(!tag->key || !tag->value) {
    printf("incomplete tag key/value %s/%s\n", tag->key, tag->value);
    gpx_tags_free(tag);
    return NULL;
  }

  for (cur_node = a_node->children; cur_node; cur_node = cur_node->next) 
    if (cur_node->type == XML_ELEMENT_NODE) 
      printf("found unhandled osm/node/tag/%s\n", cur_node->name);

  return tag;
}

/* ------------------- node handling ------------------- */

static void gpx_nodes_free(node_t *node) {
  while(node) {
    node_t *next = node->next;

    /* there must not be anything left in this chain */
    g_assert(!node->map_item_chain);

    gpx_tags_free(node->tag);
    g_free(node);

    node = next;
  }
}

void gpx_node_dump(node_t *node) {
  char buf[64];
  struct tm tm;
    
  printf("Id:      %lu\n", node->id);
  printf("User:    %s\n", node->user?node->user->name:"<unspecified>");
  printf("Visible: %s\n", node->visible?"yes":"no");
  
  localtime_r(&node->time, &tm);
  strftime(buf, sizeof(buf), "%a, %d %b %Y %H:%M:%S %Z", &tm);
  printf("Time:    %s\n", buf);
  gpx_tags_dump(node->tag);
}

void gpx_nodes_dump(node_t *node) {
  printf("\nNode list:\n");
  while(node) {
    gpx_node_dump(node);
    printf("\n");
    node = node->next;
  }
}

static node_t *gpx_parse_osm_node(gpx_t *gpx, 
			  xmlDocPtr doc, xmlNode *a_node) {
  xmlNode *cur_node = NULL;

  /* allocate a new node structure */
  node_t *node = g_new0(node_t, 1);
  node->pos.lat = node->pos.lon = NAN;

  char *prop;
  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"id"))) {
    node->id = strtoul(prop, NULL, 10);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"lat"))) {
    node->pos.lat = g_ascii_strtod(prop, NULL);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"lon"))) {
    node->pos.lon = g_ascii_strtod(prop, NULL);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"user"))) {
    node->user = gpx_user(gpx, prop);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"visible"))) {
    node->visible = (strcasecmp(prop, "true") == 0);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"timestamp"))) {
    node->time = convert_iso8601(prop);
    xmlFree(prop);
  }

  /* scan for tags and attach a list of tags */
  tag_t **tag = &node->tag;
  for (cur_node = a_node->children; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)cur_node->name, "tag") == 0) {
	/* attach tag to node */
      	*tag = gpx_parse_osm_tag(gpx, doc, cur_node);
	if(*tag) tag = &((*tag)->next);
      } else
	printf("found unhandled osm/node/%s\n", cur_node->name);
    }
  }

  /* create utm coordinates */
  utm_t utm;
  utm_pos2utm(&node->pos, &utm, &gpx->bounds->center);
  node->lpos.x =   utm.easting - gpx->bounds->center.easting;
  node->lpos.y = -utm.northing + gpx->bounds->center.northing;

  return node;
}

/* ------------------- way handling ------------------- */

void gpx_node_chain_free(node_chain_t *node_chain) {
  while(node_chain) {
    g_assert(node_chain->node->ways);

    node_chain_t *next = node_chain->next;
    node_chain->node->ways++;
    g_free(node_chain);
    node_chain = next;
  }
}

static void gpx_ways_free(way_t *way) {
  while(way) {
    way_t *next = way->next;

    gpx_node_chain_free(way->node_chain);
    gpx_tags_free(way->tag);

    /* there must not be anything left in this chain */
    g_assert(!way->map_item_chain);

    g_free(way);

    way = next;
  }
}

void gpx_way_dump(way_t *way) {
  char buf[64];
  struct tm tm;

  printf("Id:      %lu\n", way->id);
  printf("User:    %s\n", way->user?way->user->name:"<unspecified>");
  printf("Visible: %s\n", way->visible?"yes":"no");
  node_chain_t *node_chain = way->node_chain;
  while(node_chain) {
    printf("  Node:  %lu\n", node_chain->node->id);
    node_chain = node_chain->next;
  }
  
  localtime_r(&way->time, &tm);
  strftime(buf, sizeof(buf), "%a, %d %b %Y %H:%M:%S %Z", &tm);
  printf("Time:    %s\n", buf);
  gpx_tags_dump(way->tag);
}

void gpx_ways_dump(way_t *way) {
  printf("\nWay list:\n");
  while(way) {
    gpx_way_dump(way);
    printf("\n");
    way = way->next;
  }
}

node_chain_t *gpx_parse_osm_way_nd(gpx_t *gpx, 
			  xmlDocPtr doc, xmlNode *a_node) {
  char *prop;

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"ref"))) {
    item_id_t id = strtoul(prop, NULL, 10);
    node_chain_t *node_chain = g_new0(node_chain_t, 1);

    /* search matching node */
    node_chain->node = gpx->node;
    while(node_chain->node && node_chain->node->id != id)
      node_chain->node = node_chain->node->next;

    if(!node_chain->node) printf("Node id %lu not found\n", id);

    if(node_chain->node) 
      node_chain->node->ways++;

    xmlFree(prop);

    return node_chain;
  }

  return NULL;
}

static way_t *gpx_parse_osm_way(gpx_t *gpx, 
			  xmlDocPtr doc, xmlNode *a_node) {
  xmlNode *cur_node = NULL;

  /* allocate a new way structure */
  way_t *way = g_new0(way_t, 1);

  char *prop;
  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"id"))) {
    way->id = strtoul(prop, NULL, 10);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"user"))) {
    way->user = gpx_user(gpx, prop);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"visible"))) {
    way->visible = (strcasecmp(prop, "true") == 0);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"timestamp"))) {
    way->time = convert_iso8601(prop);
    xmlFree(prop);
  }

  /* scan for tags/nodes and attach their lists */
  tag_t **tag = &way->tag;
  node_chain_t **node_chain = &way->node_chain;

  for (cur_node = a_node->children; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)cur_node->name, "tag") == 0) {
	/* attach tag to node */
      	*tag = gpx_parse_osm_tag(gpx, doc, cur_node);
	if(*tag) tag = &((*tag)->next);
      } else if(strcasecmp((char*)cur_node->name, "nd") == 0) {
	*node_chain = gpx_parse_osm_way_nd(gpx, doc, cur_node);
	if(*node_chain) {
	  way->nodes++;
	  node_chain = &((*node_chain)->next);
	}
      } else
	printf("found unhandled osm/node/%s\n", cur_node->name);
    }
  }

  return way;
}

/* ------------------- relation handling ------------------- */

static void gpx_relations_free(relation_t *relation) {
  while(relation) {
    relation_t *next = relation->next;

    gpx_tags_free(relation->tag);

    member_t *member = relation->member;
    while(member) {
      member_t *next = member->next;
      if(member->role) g_free(member->role);
      g_free(member);
      member = next;
    }

    g_free(relation);

    relation = next;
  }
}

void gpx_relations_dump(relation_t *relation) {
  printf("\nRelation list:\n");
  while(relation) {
    char buf[64];
    struct tm tm;

    printf("Id:      %lu\n", relation->id);
    printf("User:    %s\n", 
	   relation->user?relation->user->name:"<unspecified>");
    printf("Visible: %s\n", relation->visible?"yes":"no");

    member_t *member = relation->member;
    while(member) {
      switch(member->type) {
      case ILLEGAL:
      case NODE_ID:
      case WAY_ID:
	break;

      case NODE:
	if(member->node)
	  printf(" Member: Node, id = %lu, role = %s\n", 
		 member->node->id, member->role);
	break;

      case WAY:
	if(member->way)
	printf(" Member: Way, id = %lu, role = %s\n", 
	       member->way->id, member->role);
	break;
      }

      member = member->next;
    }
    
    localtime_r(&relation->time, &tm);
    strftime(buf, sizeof(buf), "%a, %d %b %Y %H:%M:%S %Z", &tm);
    printf("Time:    %s\n", buf);
    gpx_tags_dump(relation->tag);
    
    printf("\n");
    relation = relation->next;
  }
}

static member_t *gpx_parse_osm_relation_member(gpx_t *gpx, 
			  xmlDocPtr doc, xmlNode *a_node) {
  char *prop;
  member_t *member = g_new0(member_t, 1);
  member->type = ILLEGAL;

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"type"))) {
    if(strcasecmp(prop, "way") == 0)       member->type = WAY;
    else if(strcasecmp(prop, "node") == 0) member->type = NODE;
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"ref"))) {
    item_id_t id = strtoul(prop, NULL, 10);

    switch(member->type) {
    case ILLEGAL:
      printf("Unable to store illegal type\n");
      break;

    case WAY:
      /* search matching way */
      member->way = gpx->way;
      while(member->way && member->way->id != id)
	member->way = member->way->next;

      if(!member->way) {
	member->type = WAY_ID;
	member->id = id;
      }
      break;

    case NODE:
      /* search matching node */
      member->node = gpx->node;
      while(member->node && member->node->id != id)
	member->node = member->node->next;

      if(!member->node) {
	member->type = NODE_ID;
	member->id = id;
      }
      break;

    case WAY_ID:
    case NODE_ID:
      break;
    }

    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"role"))) {
    if(strlen(prop) > 0) member->role = g_strdup(prop);
    xmlFree(prop);
  }

  return member;
}

static relation_t *gpx_parse_osm_relation(gpx_t *gpx, 
			  xmlDocPtr doc, xmlNode *a_node) {
  xmlNode *cur_node = NULL;

  /* allocate a new relation structure */
  relation_t *relation = g_new0(relation_t, 1);

  char *prop;
  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"id"))) {
    relation->id = strtoul(prop, NULL, 10);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"user"))) {
    relation->user = gpx_user(gpx, prop);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"visible"))) {
    relation->visible = (strcasecmp(prop, "true") == 0);
    xmlFree(prop);
  }

  if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"timestamp"))) {
    relation->time = convert_iso8601(prop);
    xmlFree(prop);
  }

  /* scan for tags and attach a list of tags */
  tag_t **tag = &relation->tag;
  member_t **member = &relation->member;

  for (cur_node = a_node->children; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)cur_node->name, "tag") == 0) {
	/* attach tag to node */
      	*tag = gpx_parse_osm_tag(gpx, doc, cur_node);
	if(*tag) tag = &((*tag)->next);
      } else if(strcasecmp((char*)cur_node->name, "member") == 0) {
	*member = gpx_parse_osm_relation_member(gpx, doc, cur_node);
	if(*member) member = &((*member)->next);
      } else
	printf("found unhandled osm/node/%s\n", cur_node->name);
    }
  }

  return relation;
}

/* ----------------------- generic xml handling -------------------------- */

/* parse loc entry */
static void gpx_parse_osm(gpx_t *gpx, xmlDocPtr doc, xmlNode * a_node) {
  xmlNode *cur_node = NULL;

  for (cur_node = a_node->children; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)cur_node->name, "bounds") == 0) 
      	gpx->bounds = gpx_parse_osm_bounds(gpx, doc, cur_node);
      else if(strcasecmp((char*)cur_node->name, "node") == 0) {
	/* parse node and attach it to chain */
      	node_t *new = gpx_parse_osm_node(gpx, doc, cur_node);
	if(new) {
	  node_t **node = &gpx->node;

#ifdef GPX_SORT_ID
	  /* search chain of nodes */
	  while(*node && ((*node)->id < new->id)) 
	    node = &(*node)->next;
#endif
	  
#ifdef GPX_SORT_LAST
	  while(*node) node = &(*node)->next;
#endif

	  /* insert into chain */
	  new->next = *node;
	  *node = new;
	}
      } else if(strcasecmp((char*)cur_node->name, "way") == 0) {
	/* parse way and attach it to chain */
      	way_t *new = gpx_parse_osm_way(gpx, doc, cur_node);
	if(new) {
	  way_t **way = &gpx->way;

#ifdef GPX_SORT_ID
	  /* insert into chain */
	  while(*way && ((*way)->id < new->id))
	    way = &(*way)->next;
#endif
	  
#ifdef GPX_SORT_LAST
	  while(*way) way = &(*way)->next;
#endif

	  /* insert into chain */
	  new->next = *way;
	  *way = new;
	}
      } else if(strcasecmp((char*)cur_node->name, "relation") == 0) {
	/* parse relation and attach it to chain */
      	relation_t *new = gpx_parse_osm_relation(gpx, doc, cur_node);
	if(new) {
	  relation_t **relation = &gpx->relation;

#ifdef GPX_SORT_ID
	  /* search chain of ways */
	  while(*relation && ((*relation)->id < new->id)) 
	    relation = &(*relation)->next;
#endif
	  
#ifdef GPX_SORT_LAST
	  while(*relation) relation = &(*relation)->next;
#endif

	  /* insert into chain */
	  new->next = *relation;
	  *relation = new;
	}
      } else
	printf("found unhandled osm/%s\n", cur_node->name);
	
    }
  }
}

/* parse root element and search for "gpx" */
static gpx_t *gpx_parse_root(xmlDocPtr doc, xmlNode * a_node) {
  gpx_t *gpx;
  xmlNode *cur_node = NULL;

  /* allocate memory to hold gpx file description */
  gpx = g_new0(gpx_t, 1);

  for (cur_node = a_node; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE) {
      /* parse osm gpx file ... */
      if(strcasecmp((char*)cur_node->name, "osm") == 0) 
      	gpx_parse_osm(gpx, doc, cur_node);
      else 
	printf("found unhandled %s\n", cur_node->name);
    }
  }

  return gpx;
}

static gpx_t *gpx_parse_doc(xmlDocPtr doc) {
  gpx_t *gpx;

  /* Get the root element node */
  xmlNode *root_element = xmlDocGetRootElement(doc);

  gpx = gpx_parse_root(doc, root_element);  

  /*free the document */
  xmlFreeDoc(doc);

  /*
   * Free the global variables that may
   * have been allocated by the parser.
   */
  xmlCleanupParser();

  return gpx;
}

/* ------------------ gpx handling ----------------- */

void gpx_free(gpx_t *gpx) {
  if(!gpx) return;

  gpx_bounds_free(gpx->bounds);
  gpx_users_free(gpx->user);
  gpx_nodes_free(gpx->node);
  gpx_ways_free(gpx->way);
  gpx_relations_free(gpx->relation);
  g_free(gpx);
}

void gpx_dump(gpx_t *gpx) {
  gpx_bounds_dump(gpx->bounds);
  gpx_users_dump(gpx->user);
  gpx_nodes_dump(gpx->node);
  gpx_ways_dump(gpx->way);
  gpx_relations_dump(gpx->relation);
}

gpx_t *gpx_parse(char *filename) {
  xmlDoc *doc = NULL;

  LIBXML_TEST_VERSION;

  /* parse the file and get the DOM */
  if ((doc = xmlReadFile(filename, NULL, 0)) == NULL) {
    xmlErrorPtr	errP = xmlGetLastError();
    errorf(NULL, "While parsing \"%s\":\n\n%s", filename, errP->message);
    return NULL;
  }
  
  return gpx_parse_doc(doc); 
}

/* ------------------------- misc access functions -------------- */

char *gpx_way_get_value(way_t *way, char *key) {
  tag_t *tag = way->tag;

  while(tag) {
    if(strcasecmp(tag->key, key) == 0)
      return tag->value;

    tag = tag->next;
  }

  return NULL;
}

char *gpx_node_get_value(node_t *node, char *key) {
  tag_t *tag = node->tag;

  while(tag) {
    if(strcasecmp(tag->key, key) == 0)
      return tag->value;

    tag = tag->next;
  }

  return NULL;
}

gboolean gpx_way_has_key_or_value(way_t *way, char *str) {
  tag_t *tag = way->tag;

  while(tag) {
    if(tag->key && strcasecmp(tag->key, str) == 0)
      return TRUE;

    if(tag->value && strcasecmp(tag->value, str) == 0)
      return TRUE;

    tag = tag->next;
  }

  return FALSE;
}

gboolean gpx_node_has_tag(node_t *node) {
  tag_t *tag = node->tag;

  if(tag && strcasecmp(tag->key, "created_by") == 0)
    tag = tag->next;

  return tag != NULL;
}

/* return true if node is part of way */
gboolean gpx_node_in_way(way_t *way, node_t *node) {
  node_chain_t *node_chain = way->node_chain;
  while(node_chain) {
    if(node_chain->node == node)
      return TRUE;

    node_chain = node_chain->next;
  }
  return FALSE;
}

static void gpx_generate_tags(tag_t *tag, xmlNodePtr node) {
  while(tag) {
    /* make sure "created_by" tag contains our id */
    if(strcasecmp(tag->key, "created_by") == 0) {
      g_free(tag->value);
      tag->value = g_strdup(PACKAGE " " VERSION);
    }

    xmlNodePtr tag_node = xmlNewChild(node, NULL, BAD_CAST "tag", NULL);
    xmlNewProp(tag_node, BAD_CAST "k", BAD_CAST tag->key);
    xmlNewProp(tag_node, BAD_CAST "v", BAD_CAST tag->value);
    tag = tag->next;
  }
}

/* build xml representation for a node or way */
char *gpx_generate_xml(gpx_t *gpx, type_t type, void *item) {
  char str[32];
  xmlChar *result = NULL;
  int len = 0;

  LIBXML_TEST_VERSION;

  xmlDocPtr doc = xmlNewDoc(BAD_CAST "1.0");
  xmlNodePtr root_node = xmlNewNode(NULL, BAD_CAST "osm");
  xmlNewProp(root_node, BAD_CAST "version", BAD_CAST "0.5");
  xmlNewProp(root_node, BAD_CAST "generator", BAD_CAST PACKAGE " V" VERSION);
  xmlDocSetRootElement(doc, root_node);

  switch(type) {
  case NODE: 
    {
      node_t *node = (node_t*)item;
      xmlNodePtr node_node = xmlNewChild(root_node, NULL, 
					 BAD_CAST "node", NULL);
      snprintf(str, sizeof(str), "%u", (unsigned)node->id);
      xmlNewProp(node_node, BAD_CAST "id", BAD_CAST str);
      g_ascii_dtostr(str, sizeof(str), node->pos.lat);
      xmlNewProp(node_node, BAD_CAST "lat", BAD_CAST str);
      g_ascii_dtostr(str, sizeof(str), node->pos.lon);
      xmlNewProp(node_node, BAD_CAST "lon", BAD_CAST str);    
      gpx_generate_tags(node->tag, node_node);
    }
    break;

  case WAY: 
    {
      way_t *way = (way_t*)item;
      xmlNodePtr way_node = xmlNewChild(root_node, NULL, BAD_CAST "way", NULL);
      snprintf(str, sizeof(str), "%u", (unsigned)way->id);
      xmlNewProp(way_node, BAD_CAST "id", BAD_CAST str);
      
      node_chain_t *node_chain = way->node_chain;
      while(node_chain) {
	xmlNodePtr nd_node = xmlNewChild(way_node, NULL, BAD_CAST "nd", NULL);
	char *str = g_strdup_printf("%ld", node_chain->node->id);
	xmlNewProp(nd_node, BAD_CAST "ref", BAD_CAST str);
	g_free(str);
	node_chain = node_chain->next;
      }
      
      gpx_generate_tags(way->tag, way_node);
    } 
    break;

  default:
    printf("neither NODE nor WAY\n");
    g_assert(0);
    break;
  }

  xmlDocDumpFormatMemoryEnc(doc, &result, &len, "UTF-8", 1);
  xmlFreeDoc(doc);
  xmlCleanupParser();

  //  puts("xml encoding result:");
  //  puts((char*)result);

  return (char*)result;
}

/* build xml representation for a node */
char *gpx_generate_xml_node(gpx_t *gpx, node_t *node) {
  return gpx_generate_xml(gpx, NODE, node);
}

/* build xml representation for a way */
char *gpx_generate_xml_way(gpx_t *gpx, way_t *way) {
  return gpx_generate_xml(gpx, WAY, way);
}

node_t *gpx_get_node_by_id(gpx_t *gpx, item_id_t id) {
  node_t *node = gpx->node;
  while(node) {
    if(node->id == id)
      return node;

    node = node->next;
  }
  
  return NULL;
}

way_t *gpx_get_way_by_id(gpx_t *gpx, item_id_t id) {
  way_t *way = gpx->way;
  while(way) {
    if(way->id == id)
      return way;

    way = way->next;
  }
  
  return NULL;
}
